
itk_transforms_c = {
    'sitkIdentity'          :   0,
    'sitkTranslation'       :   1,
    'sitkScale'             :   2,
    'sitkScaleLogarithmic'  :   3,
    'sitkEuler'             :   4,
    'sitkSimilarity'        :   5,
    'sitkQuaternionRigid'   :   6,
    'sitkVersor'            :   7,
    'sitkVersorRigid'       :   8,
    'sitkScaleSkewVersor'   :   9,
    'sitkAffine'            :   10,
    'sitkComposite'         :   11,
    'sitkDisplacementField' :   12,
    'sitkBSplineTransform'  :   13
}

image_types_c = ("original", "registered", "fused", "psf")
params_c = ("angle", "scale", "index", "channel")
